import matplotlib.pyplot as plt
import os
import pickle
import numpy as np
import seaborn as sns
import sys

import matplotlib
matplotlib.rcParams['mathtext.fontset'] = 'stix'
matplotlib.rcParams['font.family'] = 'STIXGeneral'
matplotlib.rcParams.update({'font.size': 18})

palette = sns.color_palette("colorblind")

n_runs = 10

if len(sys.argv) != 2:
    print("Usage: python plot.py <dataset>")
    sys.exit(1)


nb_points_on_plot = 100

dataset = str(sys.argv[1])
alpha = 0.01
n_rounds = 10000

show = False

seeds_list = [s for s in range(n_runs)]

norm_squared = lambda theta: np.linalg.norm(theta, axis=1, ord=2)**2
id_func = lambda x: x
avgcumsum = lambda x:  np.cumsum(x, axis=0) / np.arange(1, x.shape[0]+1)[:,None]
id_then_avg = lambda x: np.vstack((id_func(x[:len(x)//10]), avgcumsum(x[len(x)//10:])))

##################################################
## auxilliary functions


def append_to_dict(dic, data, fn):
    n_local = data["n_local"]
    theta_star = data["theta_star"]


    form = lambda x: np.array(norm_squared(fn(x - theta_star)))
    stack = lambda x, v: np.vstack((form(x), v))
    
    if n_local in dic:
        for algo in dic[n_local]:
            dic[n_local][algo] = stack(data["theta_" + algo], dic[n_local][algo])
        
    else:
        dic[n_local] = {
            "fedavg": form(data["theta_fedavg"]),
            "fedavg_rr_gamma": form(data["theta_fedavg_rr_gamma"]),
#            "fedavg_rr_local":  form(data["theta_fedavg_rr_local"]),
            "scaffold": form(data["theta_scaffold"])
        }


def plot_algo(res, color, marker, label):
    mean = np.mean(res, axis=0)
    miny = np.min(res, axis=0)
    std = np.std(res, axis=0)
    idx = np.linspace(0, len(mean)-1, len(mean), dtype=int)

    plt.plot(idx, mean[idx],
             color=color, marker=marker, label=label, markevery=len(mean)//10)
    plt.fill_between(idx,
                     np.maximum(miny, (mean-std)[idx]),
                     (mean+std)[idx],
                     color=color,
                     alpha=0.3)

def plot_all(res, heter, func_id, color_idx, n_local, dataset, show=show):
    plots_dir = "plots/plots/" + dataset + "/" + str(n_rounds) + "_" + str(alpha) + "/"
    os.makedirs(plots_dir, exist_ok=True)


    fig, ax = plt.subplots(1, 1, figsize=(4,3))                           
    res_dic = res[heter][func_id][n_local]
    
    plot_algo(res_dic["fedavg"], 
              color=palette[color_idx],
              marker="o",
              label="FedAvg")

    plot_algo(res_dic["scaffold"],
              color=palette[color_idx+1],
              marker="+",
              label="Scaffold")
    
    plot_algo(res_dic["fedavg_rr_gamma"],
              color=palette[color_idx+2],
              marker="^",
              label="Richardson-Romberg")
    
#    plot_algo(res_dic["fedavg_rr_local"],
#              color=palette[color_idx+3],
#              marker="x",
#              label="Richardson-Romberg, H")
    
    func_str = "iterates" if func_id == 0 else "averaged" if func_id == 1 else "mixed"
#    plt.title(heter + " -- " + func_str)
    plt.yscale("log")
    plt.xlabel("Communication Rounds")
    plt.ylabel("Mean Squared Error")

    if n_local == 100 and dataset == "noisy" and func_id == 2:
        plt.legend(fontsize=16)

    print("Saving", plots_dir + heter + "_" + func_str + "_" + str(n_local) + ".pdf")
    plt.savefig(plots_dir + heter + "_" + func_str + "_" + str(n_local) + ".pdf", bbox_inches="tight")
    plt.savefig(plots_dir + heter + "_" + func_str + "_" + str(n_local) + ".png", bbox_inches="tight")    

    if show:
        plt.show()


def build_results(dataset, n_rounds, alpha):
    results_dir = "plots/results/" + dataset + "/" + str(n_rounds) + "_" + str(alpha) + "/"
    
    results = {
        "homogeneous": { 0: {}, 1: {}, 2: {} },
        "heterogeneous": { 0: {}, 1: {}, 2: {} }
    }
    for heter in ["homogeneous", "heterogeneous"]:

        for seed in seeds_list:
            current_results_dir = results_dir + heter + "/" + str(seed) + "/"
    
            for func_id, func in enumerate([id_func, avgcumsum, id_then_avg]):
                

                for fpath in os.listdir(current_results_dir):
                    full_fpath = os.path.join(current_results_dir, fpath)
                
                    if os.path.isfile(full_fpath):
                        with open(full_fpath, 'rb') as f:
                            data = pickle.load(f)
                        
                            n_local = data["n_local"]
                            theta_star = data["theta_star"]
                            
                            append_to_dict(results[heter][func_id], data, func)

    return results


##################################################
# make the plots

results = build_results(dataset, n_rounds, alpha)

for heter in ["heterogeneous"]:# ["homogeneous", "heterogeneous"]:
    for n_local in [1, 10, 20, 50, 100]:
        for func_id in [0, 1, 2]:
            plot_all(results, heter=heter, func_id=func_id, n_local=n_local,
                     color_idx=0,
                     dataset=dataset)

            
                            
